# Copyright (C) 2005 Christian Limpach <Christian.Limpach@cl.cam.ac.uk>
# Copyright (C) 2005 XenSource Ltd

# This file is subject to the terms and conditions of the GNU General
# Public License.  See the file "COPYING" in the main directory of
# this archive for more details.

import os
import re
import string
import sxp
import threading
import fcntl
from struct import pack, unpack, calcsize

from xen.util.xpopen import xPopen3
import xen.util.auxbin
import xen.lowlevel.xc

import balloon
import image
from XendError import XendError, VmError
from XendLogging import log
from XendDomainInfo import DEV_MIGRATE_STEP1, DEV_MIGRATE_STEP2
from XendDomainInfo import DEV_MIGRATE_STEP3
from xen.xend.xenstore.xstransact import xstransact
from xen.xend.xenstore.xswatch import xswatch

SIGNATURE = "LinuxGuestRecord"
QEMU_SIGNATURE = "QemuDeviceModelRecord"
dm_batch = 512
XC_SAVE = "xc_save"
XC_RESTORE = "xc_restore"


sizeof_int = calcsize("i")
sizeof_unsigned_int = calcsize("I")
sizeof_unsigned_long = calcsize("L")


xc = xen.lowlevel.xc.xc()


def write_exact(fd, buf, errmsg):
    if os.write(fd, buf) != len(buf):
        raise XendError(errmsg)


def read_exact(fd, size, errmsg):
    buf  = '' 
    while size != 0: 
        readstr = os.read(fd, size)
        if not len(readstr):
            log.error("read_exact: EOF trying to read %d (buf='%s')" % \
                      (size, buf))
            raise XendError(errmsg)
        size = size - len(readstr)
        buf  = buf + readstr
    return buf

def get_avail_memory():
    """Get available memory for new guest creation (in KiB, for restore)"""
    from xen.xend import XendRoot

    # First get total memory in KiB
    xc = xen.lowlevel.xc.xc()
    info = xc.domain_getinfo()
    total_mem  = xc.physinfo()['total_memory']
    del xc

    # Count memory of all running guests in KiB
    mem_used = 0L
    for x in info:
        if x['domid'] != 0:
            # If blocked & paused & !paused & no online_vcpus we most
            # probably migrate so we get maxmem instead (since mem_kb
            # is having currently transferred amount of memory so we
            # cannot use it for good calculations)
            if (x['blocked'] == 1 and x['paused'] == 1 and
                x['running'] == 0 and x['online_vcpus'] == 0):
                log.debug("Domain %d seems to be restoring" % x['domid'])
                mem_used += x['maxmem_kb']
            else:
                mem_used += x['mem_kb']

    # Get minimal memory for dom0 and convert to KiB
    min_mem = XendRoot.instance().get_dom0_min_mem() * 1024

    # If dom0_min_mem is set to 0 then dom0 ballooning is disabled so
    # we have to use the current dom0 allocation instead
    if min_mem == 0:
        min_mem = balloon.get_dom0_current_alloc()
        log.debug("Dom0 ballooning is disabled, using current allocation (%s MiB) instead" %
                   str(min_mem / 1024) )

    # Calculate the total amount of memory that can be used by guests
    return total_mem - mem_used - min_mem

def check_for_enough_mem(val):
    mem_avail = get_avail_memory() / 1024
    log.debug("Available memory: %s MiB, guest requires: %s MiB" % (mem_avail, val))
    return int(mem_avail) > int(val)

def check_for_restore(cfg):
    for item in cfg:
        if (type(item) == list):
            # Check for enough memory to create the guest
            if item[0] == 'memory':
                if not check_for_enough_mem(item[1]):
                    raise VmError('Host machine doesn\'t have enough memory to restore the guest')

def check_is_hvm(cfg):
    for item in cfg:
        if (type(item) == list):
           if item[0] == 'image':
               return type(item[1]) == list and item[1][0] == 'hvm'

def change_hvm_readonly_ide_disks(cfg):
    for item in cfg:
        if (type(item) == list):
            if item[0] == 'device':
                # We predefine it to False for sure
                change = False
                if type(item[1]) == list and item[1][0] in ('tap', 'vbd'):
                    for p in item[1]:
                        if (type(p) != str):
                            if p[0] == 'dev':
                                change = ((p[1].find('hd') >= 0) and
                                          (p[1].find('cdrom') == -1))
                        if p[0] == 'mode' and change:
                            if p[1] == 'r':
                                p[1] = 'w'
    return cfg

def check_diskspace(location,dominfo):
    memory = dominfo.getMemoryTarget()
    # Need 1 MB memory for each 512MB of maxmem
    reserved = dominfo.getMemoryMaximum() / (1024*512) 
    s = os.statvfs(location)
    fs = (s.f_bsize * s.f_bavail) / 1024 # Needed diskspace in KB
    if fs < memory + (reserved + 40) * 1024:  # Using 40M reserve
        raise XendError("Not enough disk space to save image")

def save(fd, dominfo, network, live, dst, checkpoint=False, sock=None):
    write_exact(fd, SIGNATURE, "could not write guest state file: signature")

    config = sxp.to_string(dominfo.sxpr())

    domain_name = dominfo.getName()

    # Check for enough space to store image
    if network == False:
       check_diskspace(dst,dominfo)    

    # Check whether we can save/migrate the domain devices before start
    try:
        dominfo.checkMigrateDevices(network, dst, DEV_MIGRATE_STEP1, domain_name)
    except Exception, e:
        log.debug("Error when saving domain %s (%s): %s" % (domain_name, dominfo.domid, str(e)))
        raise e

    # Rename the domain temporarily, so that we don't get a name clash if this
    # domain is migrating (live or non-live) to the local host.  Doing such a
    # thing is useful for debugging.
    dominfo.setName('migrating-' + domain_name)

    # We should preset do_resume flag to False since we should try resume only
    # after the domain is being saved because otherwise it will return errors
    do_resume = False
    try:
        dominfo.migrateDevices(network, dst, DEV_MIGRATE_STEP1, domain_name)

        write_exact(fd, pack("!i", len(config)),
                    "could not write guest state file: config len")
        write_exact(fd, config, "could not write guest state file: config")

        hvm = dominfo.is_hvm()

        # xc_save takes three customization parameters: maxit, max_f, and
        # flags the last controls whether or not save is 'live', while the
        # first two further customize behaviour when 'live' save is
        # enabled. Passing "0" simply uses the defaults compiled into
        # libxenguest; see the comments and/or code in xc_linux_save() for
        # more information.
        cmd = [xen.util.auxbin.pathTo(XC_SAVE), str(fd),
               str(dominfo.getDomid()), "0", "0", 
               str(int(live) | (int(hvm) << 2)) ]
        log.debug("[xc_save]: %s", string.join(cmd))

        def saveInputHandler(line, tochild):
            log.debug("In saveInputHandler %s", line)
            if line == "suspend":
                log.debug("Suspending %d ...", dominfo.getDomid())
                dominfo.shutdown('suspend')
                dominfo.waitForShutdown()
                dominfo.migrateDevices(network, dst, DEV_MIGRATE_STEP2,
                                       domain_name)
                log.info("Domain %d suspended.", dominfo.getDomid())
                dominfo.migrateDevices(network, dst, DEV_MIGRATE_STEP3,
                                       domain_name)
                #send signal to device model for save
                if hvm:
                    log.info("_releaseDevices for hvm domain")
                    dominfo._releaseDevices(True)
                tochild.write("done\n")
                tochild.flush()
                log.debug('Written done')

        # By running forkHelper() cmd is being issued so when this fails we should resume
        do_resume = True
        forkHelper(cmd, fd, saveInputHandler, False, dominfo)

        # put qemu device model state
        if hvm:
            write_exact(fd, QEMU_SIGNATURE, "could not write qemu signature")
            qemu_fd = os.open("/var/lib/xen/qemu-save-%d.img" % dominfo.getDomid(), os.O_RDONLY)
            while True:
                buf = os.read(qemu_fd, dm_batch)
                if len(buf):
                    write_exact(fd, buf, "could not write device model state")
                else:
                    break
            os.close(qemu_fd)
            os.remove("/var/lib/xen/qemu-save-%d.img" % dominfo.getDomid())

        if checkpoint:
            dominfo.resumeDomain()
        else:
            dominfo._releaseDevices()
            if live and sock != None:
                sock.close()
            dominfo.destroy()

    except Exception, exn:
        log.exception("Save failed on domain %s (%s).", domain_name,
                      dominfo.getDomid())

        if do_resume:
            dominfo.resumeDomain()
            log.debug("XendCheckpoint.save: resumeDomain")
            dominfo.reconnectBackends()

        try:
            dominfo.setName(domain_name, False)
        except:
            log.exception("Failed to reset the migrating domain's name")

        raise exn

# This method checks for disk image files existence and if it doesn't exist error is returned
def check_file_exists(cfg):
    for x in cfg:
        if (type(x) != str):
            if x[0] == 'device':
                if x[1][0] in ('tap', 'vbd'):
                    for p in x[1]:
                        if (type(p) != str):
                            if p[0] == 'uname':
                                idx = p[1].find('/')
                                if idx >= 0:
                                    fn = p[1][idx:]
                                    if not os.path.exists(fn):
                                        raise XendError("Disk image %s doesn't exist" % fn)
    return True

def restore(xd, fd, paused=False, relocating=False):
    signature = read_exact(fd, len(SIGNATURE),
        "not a valid guest state file: signature read")
    if signature != SIGNATURE:
        raise XendError("not a valid guest state file: found '%s'" %
                        signature)

    l = read_exact(fd, sizeof_int,
                   "not a valid guest state file: config size read")
    vmconfig_size = unpack("!i", l)[0]
    vmconfig_buf = read_exact(fd, vmconfig_size,
        "not a valid guest state file: config read")

    p = sxp.Parser()
    p.input(vmconfig_buf)
    if not p.ready:
        raise XendError("not a valid guest state file: config parse")

    vmconfig = p.get_val()
    check_file_exists(vmconfig)

    # We check for whether we can restore the guest (e.g. there's enough memory
    # available for restore)
    check_for_restore(vmconfig)

    # If we're trying to restore an HVM guest that is using the read-only IDE
    # disks we silentry change those disks to read-write disks.
    if check_is_hvm(vmconfig):
        vmconfig = change_hvm_readonly_ide_disks(vmconfig)

    dominfo = xd.restore_(vmconfig, relocating)

    store_port   = dominfo.getStorePort()
    console_port = dominfo.getConsolePort()

    assert store_port
    assert console_port

    # if hvm, pass mem size to calculate the store_mfn
    is_hvm = dominfo.is_hvm()
    if is_hvm:
        pae  = int(sxp.child_value(dominfo.image, 'pae', 0))
        apic = int(sxp.child_value(dominfo.image, 'apic', 0))

        log.info("restore hvm domain %d, apic=%d, pae=%d",
                 dominfo.domid, apic, pae)
    else:
        apic = 0
        pae  = 0

    try:
        restore_image = image.create(dominfo, dominfo.info['image'],
                                     dominfo.info['device'])
        memory = restore_image.getRequiredAvailableMemory(
            dominfo.info['memory'] * 1024)
        maxmem = restore_image.getRequiredAvailableMemory(
            dominfo.info['maxmem'] * 1024)
        shadow = restore_image.getRequiredShadowMemory(
            dominfo.info['shadow_memory'] * 1024,
            dominfo.info['maxmem'] * 1024)

        log.debug("restore:shadow=0x%x, _static_max=0x%x, _static_min=0x%x, ",
                  dominfo.info['shadow_memory'],
                  dominfo.info['maxmem'],
                  dominfo.info['memory'])

        # Round shadow up to a multiple of a MiB, as shadow_mem_control
        # takes MiB and we must not round down and end up under-providing.
        shadow = ((shadow + 1023) / 1024) * 1024

        # set memory limit
        xc.domain_setmaxmem(dominfo.getDomid(), maxmem)

        vtd_mem = 0
        # On VT-d enabled host, let's reserve extra memory for hvm guest.
        if dominfo.is_hvm() and xc.test_assign_device(0, 0, 0, 0, 0) != -1:
           # Reserve 1 page per MiB of RAM for separate VT-d page table.
           vtd_mem = 4 * (maxmem / 1024)
           # Round vtd_mem up to a multiple of a MiB.
           vtd_mem = ((vtd_mem + 1023) / 1024) * 1024

        balloon.free(memory + shadow + vtd_mem)

        shadow_cur = xc.shadow_mem_control(dominfo.getDomid(), shadow / 1024)
        dominfo.info['shadow_memory'] = shadow_cur

        cmd = map(str, [xen.util.auxbin.pathTo(XC_RESTORE),
                        fd, dominfo.getDomid(),
                        store_port, console_port, int(is_hvm), pae, apic])
        log.debug("[xc_restore]: %s", string.join(cmd))

        handler = RestoreInputHandler()

        forkHelper(cmd, fd, handler.handler, True)

        # We don't want to pass this fd to any other children -- we 
        # might need to recover the disk space that backs it.
        try:
            flags = fcntl.fcntl(fd, fcntl.F_GETFD)
            flags |= fcntl.FD_CLOEXEC
            fcntl.fcntl(fd, fcntl.F_SETFD, flags)
        except:
            pass

        if handler.store_mfn is None:
            raise XendError('Could not read store MFN')

        if not is_hvm and handler.console_mfn is None:
            raise XendError('Could not read console MFN')        

        # get qemu state and create a tmp file for dm restore
        if is_hvm:
            qemu_signature = read_exact(fd, len(QEMU_SIGNATURE),
                                        "invalid device model signature read")
            if qemu_signature != QEMU_SIGNATURE:
                raise XendError("not a valid device model state: found '%s'" %
                                qemu_signature)
            qemu_fd = os.open("/var/lib/xen/qemu-save-%d.img" % dominfo.getDomid(),
                              os.O_WRONLY | os.O_CREAT | os.O_TRUNC)
            while True:
                buf = os.read(fd, dm_batch)
                if len(buf):
                    write_exact(qemu_fd, buf,
                                "could not write dm state to tmp file")
                else:
                    break
            os.close(qemu_fd)


        os.read(fd, 1)           # Wait for source to close connection
        
        dominfo.completeRestore(handler.store_mfn, handler.console_mfn)

        #
        # We shouldn't hold the domains_lock over a waitForDevices
        # As this function sometime gets called holding this lock,
        # we must release it and re-acquire it appropriately
        #
        from xen.xend import XendDomain

        lock = True;
        try:
            XendDomain.instance().domains_lock.release()
        except:
            lock = False;

        try:
            dominfo.waitForDevices() # Wait for backends to set up
        except Exception, exn:
            log.exception(exn)

        if lock:
            XendDomain.instance().domains_lock.acquire()

        if not paused:
            dominfo.unpause()

        return dominfo
    except:
        dominfo.destroy()
        raise


class RestoreInputHandler:
    def __init__(self):
        self.store_mfn = None
        self.console_mfn = None


    def handler(self, line, _):
        m = re.match(r"^(store-mfn) (\d+)$", line)
        if m:
            self.store_mfn = int(m.group(2))
        else:
            m = re.match(r"^(console-mfn) (\d+)$", line)
            if m:
                self.console_mfn = int(m.group(2))


def forkHelper(cmd, fd, inputHandler, closeToChild, dominfo=None):
    child = xPopen3(cmd, True, -1, [fd, xc.handle()])

    if closeToChild:
        child.tochild.close()

    thread = threading.Thread(target = slurp, args = (child.childerr,))
    thread.start()

    try:
        try:
            while 1:
                line = child.fromchild.readline()
                if line == "":
                    break
                else:
                    line = line.rstrip()
                    log.debug('%s', line)
                    inputHandler(line, child.tochild)

        except IOError, exn:
            raise XendError('Error reading from child process for %s: %s' %
                            (cmd, exn))
    finally:
        child.fromchild.close()
        if not closeToChild:
            child.tochild.close()
        thread.join()
        child.childerr.close()
        status = child.wait()

    if status >> 8 == 127:
        raise XendError("%s failed: popen failed" % string.join(cmd))
    elif status != 0:
        # For failed save or migrate we need to restore the original
        # name. Migrate is using the same code path like save except
        # it is saving to the socket descriptor instead of file. On
        # the other end the guest is restored in case of migration.
        if cmd[0].endswith("xc_save"):
            if dominfo is not None:
                if dominfo.getName().startswith("migrating-"):
                    dominfo.setName( dominfo.getName()[10:] )
                    log.debug("Domain %d renamed back to %s" %
                                 (dominfo.domid, dominfo.getName()) )

                # For HVM guest we also need to create a new device model
                # since the old one no longer exists. Also we need to
                # remove shutdown control entry not to confuse device
                # model process, i.e. qemu-dm
                if dominfo.is_hvm():
                    xstransact.Remove("/local/domain/%s/control/shutdown"
                                                          % dominfo.domid)
                    dominfo.image.createDeviceModel(True)

        raise XendError("%s failed" % string.join(cmd))


def slurp(infile):
    while 1:
        line = infile.readline()
        if line == "":
            break
        else:
            line = line.strip()
            m = re.match(r"^ERROR: (.*)", line)
            if m is None:
                log.info('%s', line)
            else:
                log.error('%s', m.group(1))
